"""Data access object for timetable"""
import os
from pathlib import Path
from dataclasses import dataclass

from loguru import logger
import joblib

from pyraptor.util import mkdir_if_not_exists


@dataclass
class Timetable:
    """Timetable data"""

    stations = None
    stops = None
    trips = None
    trip_stop_times = None


def read_timetable(input_folder: str) -> Timetable:
    """
    Read the timetable data from the cache directory
    """

    if not os.path.exists(input_folder):
        raise IOError(
            "Optimized timetable not found. Run `pyraptor/gtfs/timetable.py` first to create timetable."
        )

    logger.debug("Using cached optimized datastructures")

    def load_joblib(name):
        logger.debug(f"Loading '{name}'")
        with open(Path(input_folder, f"{name}.pcl"), "rb") as handle:
            return joblib.load(handle)

    timetable = Timetable()

    timetable.stations = load_joblib("stations")
    timetable.stops = load_joblib("stops")
    timetable.trips = load_joblib("trips")
    timetable.trip_stop_times = load_joblib("trip_stop_times")

    return timetable


def write_timetable(output_folder: str, timetable: Timetable) -> None:
    """
    Write the timetable to output directory
    """

    def write_joblib(state, name):
        with open(Path(output_folder, f"{name}.pcl"), "wb") as handle:
            joblib.dump(state, handle)

    logger.info("Write optimized timetable to output directory")

    mkdir_if_not_exists(output_folder)

    write_joblib(timetable.stations, "stations")
    write_joblib(timetable.stops, "stops")
    write_joblib(timetable.trips, "trips")
    write_joblib(timetable.trip_stop_times, "trip_stop_times")



"""Save results from RAPTOR algorithm"""
import os
from datetime import datetime

import pandas as pd
from loguru import logger

from pyraptor.util import sec2str, mkdir_if_not_exists


def write_results(output_folder, timetable, bag_k, evaluations) -> None:
    """
    Export results to a CSV file with stations and traveltimes per iteration.
    """

    mkdir_if_not_exists(output_folder)
    now = datetime.now()

    # Traveltime per round
    rows = []

    for round_k in list(bag_k.keys()):
        legs = bag_k[round_k]
        destination = 0

        for tt in legs:
            stop = timetable.stops.set_index.get(destination, None)

            if stop:
                name = stop.station.name
                platform = stop.platform_code

                rows.append(
                    dict(
                        round=str(round_k),
                        stop_id=str(destination),
                        stop_name=str(name),
                        platform_code=str(platform),
                        travel_time=str(tt[0]),
                    )
                )
            destination = destination + 1
    df = pd.DataFrame(rows)

    df = (
        df[["round", "stop_name", "travel_time"]]
        .groupby(["round", "stop_name"])
        .min()
        .sort_values(by=["stop_name", "round"])
    )
    df.travel_time = df.travel_time.apply(lambda x: sec2str(int(x)))

    filename1 = os.path.join(
        output_folder, "{date:%Y%m%d_%H%M%S}_traveltime.csv".format(date=now)
    )
    logger.debug("Write results to {}".format(filename1))
    df.to_csv(filename1)

    # Last legs
    rounds = sorted(list(bag_k.keys()))
    bag = bag_k[rounds[-1]]

    rows = []
    for b in bag:
        frm = b[0]
        via = b[1]
        to = b[2]
        rows.append(
            dict(
                from_id=str(frm),
                trip_id=str(via),
                stop_id=str(to),
            )
        )
    df2 = pd.DataFrame(rows)

    filename2 = os.path.join(
        output_folder, "{date:%Y%m%d_%H%M%S}_last_legs.csv".format(date=now)
    )
    logger.debug("Write results to {}".format(filename2))
    df2.to_csv(filename2, index=False)

    # Evaluations
    # (k, start_stop, trip, arrival_trip_stop_time)
    rows = []

    for val in evaluations:
        k = val[0]
        start_stop = val[1]
        trip = val[2]
        arrival_stop_time = val[3]

        rows.append(
            dict(
                k=k,
                start_stop=start_stop.id,
                trip=trip.id,
                to=arrival_stop_time.stop.id,
                trip_id=trip.id,
                trip_number=trip.hint,
                arrival=arrival_stop_time.dts_arr,
                from_name=start_stop.station.name,
                from_platform=start_stop.platform_code,
                to_name=arrival_stop_time.stop.station.name,
                to_platform=arrival_stop_time.stop.platform_code,
            )
        )
    df3 = pd.DataFrame(rows)
    df3.arrival = df3.arrival.apply(sec2str)

    filename3 = os.path.join(
        output_folder,
        "{date:%Y%m%d_%H%M%S}_evaluations.csv".format(date=now),
    )
    logger.debug("Write results to {}".format(filename3))
    df3.to_csv(filename3, index=False)



"""Utility functions"""
import os
import numpy as np


SAVE_RESULTS = True
TRANSFER_COST = 3 * 60  # Default transfer time is 3 minutes
T3M = 3 * 60
T1H = 1 * 60 * 60
T6H = 6 * 60 * 60
T24H = 24 * 60 * 60


def mkdir_if_not_exists(name):
    """Create directory if not exists"""
    if not os.path.exists(name):
        os.makedirs(name)


def str2sec(time_str):
    """
    Convert hh:mm:ss to seconds since midnight
    :param time_str: String in format hh:mm:ss
    """
    split_time = time_str.strip().split(":")
    if len(split_time) == 3:
        # Has seconds
        hours, minutes, seconds = split_time
        return int(hours) * 3600 + int(minutes) * 60 + int(seconds)
    minutes, seconds = split_time
    return int(minutes) * 60 + int(seconds)


def sec2str(scnds, show_sec=False):
    """
    Convert hh:mm:ss to seconds since midnight

    :param show_sec: only show :ss if True
    :param scnds: Seconds to translate to hh:mm:ss
    """
    scnds = np.round(scnds)
    hours = int(scnds / 3600)
    minutes = int((scnds % 3600) / 60)
    seconds = int(scnds % 60)
    return (
        "{:02d}:{:02d}:{:02d}".format(hours, minutes, seconds)
        if show_sec
        else "{:02d}:{:02d}".format(hours, minutes)
    )



"""Run range query on RAPTOR algorithm"""
import argparse
from time import perf_counter

import numpy as np
from loguru import logger

from pyraptor.dao.timetable import Timetable, read_timetable
from pyraptor.model.raptor import (
    traverse_trips,
    add_transfer_time,
    final_destination,
    reconstruct_journey,
    is_dominated,
    print_journey,
)
from pyraptor.util import (
    str2sec,
    sec2str,
    T24H,
)


def parse_arguments():
    """Parse arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="output/optimized_timetable",
        help="Input directory",
    )
    parser.add_argument(
        "-or",
        "--origin",
        type=str,
        default="Hertogenbosch ('s)",
        help="Origin station of the journey",
    )
    parser.add_argument(
        "-d",
        "--destination",
        type=str,
        default="Rotterdam Centraal",
        help="Destination station of the journey for logging purposes",
    )
    parser.add_argument(
        "-st",
        "--starttime",
        type=str,
        default="08:00:00",
        help="Start departure time (hh:mm:ss)",
    )
    parser.add_argument(
        "-et",
        "--endtime",
        type=str,
        default="08:30:00",
        help="End departure time (hh:mm:ss)",
    )
    parser.add_argument(
        "-r",
        "--rounds",
        type=int,
        default=4,
        help="Number of rounds to execute the RAPTOR algorithm",
    )
    arguments = parser.parse_args()

    return arguments


def main(
    input_folder: str,
    origin_station: str,
    destination_station: str,
    departure_start_time: str,
    departure_end_time: str,
    rounds: int,
):
    """Run RAPTOR algorithm"""

    logger.debug("Input directory      : {}", input_folder)
    logger.debug("Origin station       : {}", origin_station)
    logger.debug("Destination station  : {}", destination_station)
    logger.debug("Departure start time : {}", departure_start_time)
    logger.debug("Departure end time   : {}", departure_end_time)
    logger.debug("Rounds               : {}", str(rounds))

    timetable = read_timetable(input_folder)

    logger.info("Calculating network from : {}".format(origin_station))

    # Departure time seconds for time range
    dep_secs_min = str2sec(departure_start_time)
    dep_secs_max = str2sec(departure_end_time)
    logger.debug(
        "Departure time range (s.)  : ({}, {})".format(dep_secs_min, dep_secs_max)
    )

    # Find route between two stations for time range, i.e. Range Query
    # traveltime, final_dest, stop_bag
    start = perf_counter()
    labels = perform_recursive_raptor(
        timetable,
        origin_station,
        dep_secs_min,
        dep_secs_max,
        rounds,
    )

    # All destinations are present in labels, so this is only for logging purposes
    print_journeys(timetable, labels, destination_station=destination_station)

    logger.info(
        "RAPTOR Algorithm executed in {:.4f} seconds".format(perf_counter() - start)
    )


def perform_recursive_raptor(
    timetable: Timetable,
    origin_station: str,
    dep_secs_min: int,
    dep_secs_max: int,
    rounds: int,
):
    """
    Perform the RAPTOR algorithm for a range query.
    """

    # Determine stop IDs for origins and destinations
    from_stops = timetable.stations.get_stops(origin_station)
    destination_stop_ids = {
        st.name: timetable.stations.get_stops(st.name) for st in timetable.stations
    }
    destination_stop_ids.pop(origin_station, None)

    # Find all trips leaving from stops within time range
    potential_trip_stop_times = timetable.trip_stop_times.get_trip_stop_times_in_range(
        from_stops, dep_secs_min, dep_secs_max
    )
    potential_dep_secs = sorted(
        list(set([tst.dts_dep for tst in potential_trip_stop_times])), reverse=True
    )

    logger.info(
        "Potential departure times : {}".format(
            [sec2str(x) for x in potential_dep_secs]
        )
    )

    labels = {station_name: [] for station_name, _ in destination_stop_ids.items()}
    last_round_labels = {
        station_name: None for station_name, _ in destination_stop_ids.items()
    }

    for dep_index, dep_secs in enumerate(potential_dep_secs):
        logger.info(f"Processing {dep_index} / {len(potential_dep_secs)}")
        logger.info("Analyzing best journey for departure time {}".format(dep_secs))

        # Initialize lookup with start node taking 0 seconds to reach
        # Bag contains per stop (travel_time, trip_id, previous_stop), trip_id is 0 in case of a transfer
        number_stops = len(timetable.stops) + 1
        bag = np.full(
            shape=(number_stops, 3),
            fill_value=(T24H, 0, -1),
            dtype=np.dtype(np.int32, np.int32, np.int32),
        )

        # Add origin stops to bag
        logger.debug("Starting from Stop IDs: {}".format(str(from_stops)))

        new_stops = []
        for from_stop in from_stops:
            bag[from_stop.index] = (0, 0, 0)
            new_stops.append(from_stop)

        # Run Round-Based Algorithm
        bag_k = {}
        evaluations = []

        for k in range(1, rounds + 1):

            logger.info("Analyzing possibilities round {}".format(k))

            # Get list of stops to evaluate in the process
            logger.debug("Stops to evaluate count: {}".format(len(new_stops)))

            # Update time to stops calculated based on stops reachable
            new_stops_travel, bag, evaluations = traverse_trips(
                timetable, evaluations, k, new_stops, bag, dep_secs
            )
            logger.debug("{} reachable stops added".format(len(new_stops_travel)))

            # Now add footpath transfers and update
            new_stops_transfer, bag = add_transfer_time(new_stops_travel, bag)
            logger.debug("{} transferable stops added".format(len(new_stops_transfer)))

            new_stops = set(new_stops_travel).union(new_stops_transfer)
            logger.debug("{} stops to evaluate in next round".format(len(new_stops)))

            # Store the results for this round
            bag_k[k] = np.copy(bag)

        # Determine the best destination ID, destination is a platform
        for destination_station_name, to_stops in destination_stop_ids.items():
            dest_id = final_destination(to_stops, bag)
            if dest_id != 0:
                journey = reconstruct_journey(dest_id, bag)
                last_round_journey = last_round_labels[destination_station_name]

                last_round_labels[destination_station_name] = journey

                if not is_dominated(timetable, last_round_journey, journey):
                    labels[destination_station_name].append(journey)

    return labels


def print_journeys(timetable: Timetable, labels, destination_station: str):
    """Print journeys"""
    logger.info("---")
    logger.info(f"Destination station {destination_station}")
    for journey in labels[destination_station][::-1]:
        print_journey(timetable, journey)
        logger.info("---")


if __name__ == "__main__":
    args = parse_arguments()
    main(
        args.input,
        args.origin,
        args.destination,
        args.starttime,
        args.endtime,
        args.rounds,
    )



"""Run RAPTOR algorithm"""
from typing import List, Tuple

from loguru import logger

from pyraptor.dao.timetable import Timetable
from pyraptor.model.datatypes import Stop, Trip, TripStopTime
from pyraptor.util import (
    str2sec,
    sec2str,
    SAVE_RESULTS,
    T1H,
    T6H,
    T24H,
    TRANSFER_COST,
)


def traverse_trips(
    timetable: Timetable, evaluations: List, k: int, ids: List, bag, dep_secs: int
) -> Tuple:
    """
    Iterator through the stops reachable and add all new reachable stops
    by following all trips from the reached stations. Trips are only followed
    in the direction of travel and beyond already added points.

    :param k: current round
    :param ids: current stops reached
    :param bag: numpy array with info over reached stops
    :param dep_secs: Departure time in seconds
    """

    new_stops = []

    n_evaluations = 0
    n_improvements = 0

    for start_stop in ids:

        # how long it took to get to the stop so far (0 for start node)
        time_sofar = bag[start_stop.index][0]

        # get list of all trips associated with this stop
        trips = get_trip_ids_for_stop(timetable, start_stop, dep_secs + time_sofar, T1H)

        for (trip, current_stopidx) in trips:

            # get all following stop times for this trip
            stop_times = trip.get_next_trip_stop_times(current_stopidx)

            # for all following stops, calculate time to reach
            for arrive_stop_time in stop_times:
                n_evaluations += 1

                if SAVE_RESULTS:
                    evaluations.append((k, start_stop, trip, arrive_stop_time))

                # time to reach is diff from start time to arrival (plus any baseline cost)
                arrive_time_adjusted = arrive_stop_time.dts_arr - dep_secs

                # only update if does not exist yet or is faster
                old_value = bag[arrive_stop_time.stop.index][0]
                if arrive_time_adjusted < old_value:
                    n_improvements += 1
                    bag[arrive_stop_time.stop.index] = (
                        arrive_time_adjusted,
                        trip.id,
                        start_stop.index,
                    )
                    new_stops.append(arrive_stop_time.stop)

    logger.debug("- Evaluations    : {}".format(n_evaluations))
    logger.debug("- Improvements   : {}".format(n_improvements))

    return new_stops, bag, evaluations


def get_trip_ids_for_stop(
    timetable: Timetable,
    stop_id: Stop,
    dep_secs: int,
    forward: int = T6H,
) -> List[Tuple[Trip, int]]:
    """
    Takes a stop and departure time and get associated trip ids.
    The forward parameter limits the time frame starting at the departure time.
    Default framesize is 60 minutes.
    Times are specified in seconds since midnight.

    :param stop_id: Stop
    :param dep_time: Departure time
    :param forward: Period forward limitimg trips
    """
    trip_stop_times = timetable.trip_stop_times.get_trip_stop_times_for_stop(
        stop_id, dep_secs, forward
    )
    trips = list(set([(tst.trip, tst.stopidx) for tst in trip_stop_times]))
    return trips


def add_transfer_time(stops: List[Stop], bag) -> Tuple:
    """Add transfers between platforms."""

    # TODO Implement transfer station boolean

    new_stops = []

    # add in transfers to other platforms
    for stop in stops:

        station = stop.station
        other_station_stops = [st for st in station.stops if st != stop]

        time_sofar = bag[stop.index][0]
        for arrive_stop in other_station_stops:
            arrive_time_adjusted = time_sofar + get_transfer_time(
                stop, arrive_stop, time_sofar, 0
            )
            old_value = bag[arrive_stop.index][0]
            if arrive_time_adjusted < old_value:
                bag[arrive_stop.index] = (arrive_time_adjusted, 0, stop.index)
                new_stops.append(arrive_stop)

    return new_stops, bag


def get_transfer_time(stop_from: int, stop_to: int, time_sec: int, dow: int) -> int:
    """
    Calculate the transfer time from a stop to another stop (usually two platforms at one station
    :param stop_from: Origin platform
    :param stop_to: Destination platform
    :param time_sec: Time of day (seconds since midnight)
    :param dow: day of week (Monday = 0, Tuesday = 1, ...)
    """
    return TRANSFER_COST


def is_dominated(timetable: Timetable, original_journey, new_journey) -> bool:
    """Check if new journey is dominated by another journey"""
    # First journey
    if not original_journey:
        return False

    # No improvement
    if original_journey == new_journey:
        return True

    def depart(jrny) -> int:
        depart_leg = jrny[0] if jrny[0][1] != 0 else jrny[1]
        depart_stop = timetable.stops.get_by_index(depart_leg[0])
        depart_trip = timetable.trips.set_idx[depart_leg[1]]
        depart_stop_time = [
            st for st in depart_trip.stop_times if st.stop == depart_stop
        ][0]
        return depart_stop_time.dts_dep

    def arrival(jrny) -> int:
        arrival_leg = jrny[-1]
        arrival_stop = timetable.stops.get_by_index(arrival_leg[2])
        arrival_trip = timetable.trips.set_idx[arrival_leg[1]]
        arrival_stop_time = [
            st for st in arrival_trip.stop_times if st.stop == arrival_stop
        ][0]
        return arrival_stop_time.dts_arr

    original_depart = depart(original_journey)
    new_depart = depart(new_journey)

    original_arrival = arrival(original_journey)
    new_arrival = arrival(new_journey)

    # Is dominated
    return (
        True
        if original_depart > new_depart and original_arrival < new_arrival
        else False
    )


def final_destination(to_stops: List[Stop], bag) -> Stop:
    """
    Find the destination ID with the shortest distance.
    Required in order to prevent adding travel time to the arrival time.

    :param to_ids:
    :param bag:
    """
    final_stop = 0
    distance = T24H
    for stop in to_stops:
        if bag[stop.index][0] < distance:
            distance = bag[stop.index][0]
            final_stop = stop
    return final_stop


def reconstruct_journey(destination: Stop, bag):
    """Construct journey for destination from values in bag."""
    jrny = []
    current = destination.index
    while current != 0:
        # bag = (travel_time, trip_id, previous_stop.index)
        # journey = (previous_stop.index, trip.id, to_stop.index)
        jrny.append((bag[current][2], bag[current][1], current))
        current = bag[current][2]
    jrny.reverse()
    return jrny


def print_journey(timetable, journey, dep_secs=None):
    """
    Print the given journey to logger info
    """
    logger.info("Journey:")

    if len(journey) == 0:
        logger.info("No journey available")
        return

    # Print all legs in journey
    # leg = (previous_stop.index, trip.id, to_stop.index)
    for leg in journey:
        if leg[1] != 0:
            # Stop and trip
            from_stop = timetable.stops.get_by_index(leg[0])
            to_stop = timetable.stops.get_by_index(leg[2])
            trip = timetable.trips.set_idx[leg[1]]

            dep = [tst.dts_dep for tst in trip.stop_times if from_stop == tst.stop][0]
            arr = [tst.dts_arr for tst in trip.stop_times if to_stop == tst.stop][0]

            msg = (
                str(sec2str(dep))
                + " "
                + from_stop.station.name.ljust(20)
                + "(p. "
                + str(from_stop.platform_code).rjust(3)
                + ") TO "
                + str(sec2str(arr))
                + " "
                + to_stop.station.name.ljust(20)
                + "(p. "
                + str(to_stop.platform_code).rjust(3)
                + ") WITH "
                + str(trip.hint)
            )
            logger.info(msg)

    # Departure time of first leg
    depart_leg = journey[0] if journey[0][1] != 0 else journey[1]
    depart_stop = timetable.stops.get_by_index(depart_leg[0])
    depart_trip = timetable.trips.set_idx[depart_leg[1]]
    depart_stop_time = [st for st in depart_trip.stop_times if st.stop == depart_stop][
        0
    ]

    # Arrival time of last leg
    arrival_leg = journey[-1]
    arrival_stop = timetable.stops.get_by_index(arrival_leg[2])
    arrival_trip = timetable.trips.set_idx[arrival_leg[1]]
    arrival_stop_time = [
        st for st in arrival_trip.stop_times if st.stop == arrival_stop
    ][0]

    msg = "Duration : {}".format(
        sec2str(arrival_stop_time.dts_arr - depart_stop_time.dts_dep)
    )
    if dep_secs:
        msg += " ({} from request time {})".format(
            sec2str(arrival_stop_time.dts_arr - dep_secs),
            sec2str(dep_secs),
        )
    logger.info(msg)



"""Datatypes"""
from collections import defaultdict
from typing import List

import attr

from pyraptor.util import T6H


def same_type_and_id(first, second):
    """Same type and ID"""
    return type(first) is type(second) and first.id == second.id


@attr.s(repr=False, cmp=False)
class Stop:
    """Stop"""

    id = attr.ib(default=None)
    name = attr.ib(default=None)
    station = attr.ib(default=None)
    platform_code = attr.ib(default=None)
    index = attr.ib(default=attr.Factory(lambda seq=iter(range(2 ** 40)): next(seq)))

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, stop):
        return same_type_and_id(self, stop)

    def __repr__(self):
        if self.id == self.name:
            return "<Stop {}>".format(self.id)
        return "<Stop {} [{}]>".format(self.name, self.id)


class Stops:
    """Stops"""

    def __init__(self):
        self.set_idx = dict()
        self.set_index = dict()

    def __repr__(self):
        return f"Stops(n_stops={len(self.set_idx)})"

    def __getitem__(self, stop_id):
        return self.set_idx[stop_id]

    def __len__(self):
        return len(self.set_idx)

    def __iter__(self):
        return iter(self.set_idx.values())

    def get(self, stop_id):
        """Get stop"""
        if stop_id not in self.set_idx:
            raise ValueError(f"Stop ID {stop_id} not present in Stops")
        stop = self.set_idx[stop_id]
        return stop

    def get_by_index(self, stop_index):
        return self.set_index[stop_index]

    def add(self, stop):
        """Add stop"""
        if stop.id in self.set_idx:
            stop = self.set_idx[stop.id]
        else:
            self.set_idx[stop.id] = stop
            self.set_index[stop.index] = stop
        return stop


@attr.s(repr=False, cmp=False)
class Station:
    """Stop dataclass"""

    id = attr.ib(default=None)
    name = attr.ib(default=None)
    stops = attr.ib(default=attr.Factory(list))

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, stop):
        return same_type_and_id(self, stop)

    def __repr__(self):
        if self.id == self.name:
            return "<Station {}>".format(self.id)
        return "<Station {} [{}]>".format(self.name, self.id)

    def add_stop(self, stop: Stop):
        self.stops.append(stop)


class Stations:
    """Stations"""

    def __init__(self):
        self.set_idx = dict()

    def __repr__(self):
        return f"<Stations(n_stations={len(self.set_idx)})>"

    def __getitem__(self, station_id):
        return self.set_idx[station_id]

    def __len__(self):
        return len(self.set_idx)

    def __iter__(self):
        return iter(self.set_idx.values())

    def add(self, station: Station):
        if station.id in self.set_idx:
            station = self.set_idx[station.id]
        else:
            self.set_idx[station.id] = station
        return station

    def get(self, station: Station):
        if isinstance(station, Station):
            station = station.id
        if station not in self.set_idx:
            return None
        return self.set_idx[station]

    def get_stops(self, station_name):
        """Get all stop ids from station, i.e. platform stop ids belonging to station"""
        return self.set_idx[station_name].stops


@attr.s(repr=False)
class TripStopTime:
    """Trip Stop"""

    trip = attr.ib(default=attr.NOTHING)
    stopidx = attr.ib(default=attr.NOTHING)
    stop = attr.ib(default=attr.NOTHING)
    dts_arr = attr.ib(default=attr.NOTHING)
    dts_dep = attr.ib(default=attr.NOTHING)

    def __hash__(self):
        return hash((self.trip, self.stopidx))

    def __repr__(self):
        return (
            "TripStopTime(trip_id={hint}{trip_id}, stopidx={0.stopidx},"
            " stop_id={0.stop.id}, dts_arr={0.dts_arr}, dts_dep={0.dts_dep})"
        ).format(
            self,
            trip_id=self.trip.id if self.trip else None,
            hint="{}:".format(self.trip.hint)
            if self.trip and self.trip.hint
            else "",
        )


class TripStopTimes:
    """Trip Stop Times"""

    def __init__(self):
        self.set_idx = dict()
        self.stop_trip_idx = defaultdict(list)

    def __repr__(self):
        return f"Trips(n_trips={len(self.set_idx)})"

    def __getitem__(self, trip_id):
        return self.set_idx[trip_id]

    def __len__(self):
        return len(self.set_idx)

    def __iter__(self):
        return iter(self.set_idx.values())

    def add(self, trip_stop_time: TripStopTime):
        self.set_idx[(trip_stop_time.trip, trip_stop_time.stopidx)] = trip_stop_time
        self.stop_trip_idx[trip_stop_time.stop].append(trip_stop_time)

    def get_trip_stop_times_in_range(self, stops, dep_secs_min, dep_secs_max):
        """Returns all trip stop times with departure time within range"""
        in_window = [
            tst
            for tst in self
            if tst.dts_dep >= dep_secs_min
            and tst.dts_dep <= dep_secs_max
            and tst.stop in stops
        ]
        return in_window

    def get_trip_stop_times_for_stop(
        self, stop: Stop, dep_secs: int, forward: int = T6H
    ) -> List[TripStopTime]:
        """
        Takes a stop and departure time and get associated trip ids.
        The forward parameter limits the time frame starting at the departure time.
        Times are specified in seconds since midnight.

        :param stop_id: Stop
        :param dep_secs: Departure time
        :param forward: Period forward limitimg trips
        """
        trip_stop_times = self.stop_trip_idx[stop]
        in_window = [
            tst
            for tst in trip_stop_times
            if tst.dts_dep >= dep_secs and tst.dts_dep <= dep_secs + forward
        ]
        return in_window


@attr.s(repr=False, cmp=False)
class Trip:
    """Trip"""

    id = attr.ib(default=attr.Factory(lambda seq=iter(range(2 ** 40)): next(seq)))
    stop_times = attr.ib(default=attr.Factory(list))
    hint = attr.ib(default=None)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, trip):
        return same_type_and_id(self, trip)

    def __repr__(self):
        return "Trip(id={hint}:{0.id}, stop_times={stop_times}:{first_stop}-{last_stop})".format(
            self,
            stop_times=len(self.stop_times),
            first_stop=self.stop_times[0].stop.id,
            last_stop=self.stop_times[-1].stop.id,
            hint=self.hint,
        )

    def __getitem__(self, n):
        return self.stop_times[n]

    def __len__(self):
        return len(self.stop_times)

    def __iter__(self):
        return iter(self.stop_times)

    def add_stop_time(self, stop_time: TripStopTime):
        assert stop_time.dts_arr <= stop_time.dts_dep
        assert not self.stop_times or self.stop_times[-1].dts_dep <= stop_time.dts_arr
        self.stop_times.append(stop_time)

    def get_next_trip_stop_times(self, stop_idx: int):
        return [st for st in self.stop_times if st.stopidx > stop_idx]


class Trips:
    """Trips"""

    def __init__(self):
        self.set_idx = dict()

    def __repr__(self):
        return f"Trips(n_trips={len(self.set_idx)})"

    def __getitem__(self, trip_id):
        return self.set_idx[trip_id]

    def __len__(self):
        return len(self.set_idx)

    def __iter__(self):
        return iter(self.set_idx.values())

    def add(self, trip):
        assert len(trip) >= 2, "must have 2 stop times"
        self.set_idx[trip.id] = trip



"""Run query with RAPTOR algorithm"""
import argparse
from time import perf_counter

import numpy as np
from loguru import logger

from pyraptor.dao.timetable import Timetable, read_timetable
from pyraptor.dao.results import write_results
from pyraptor.model.raptor import (
    print_journey,
    traverse_trips,
    reconstruct_journey,
    final_destination,
    add_transfer_time,
)
from pyraptor.util import (
    str2sec,
    SAVE_RESULTS,
    T24H,
)


def parse_arguments():
    """Parse arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="output/optimized_timetable",
        help="Input directory",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="output/results",
        help="Input directory",
    )
    parser.add_argument(
        "-or",
        "--origin",
        type=str,
        default="Hertogenbosch ('s)",
        help="Origin station of the journey",
    )
    parser.add_argument(
        "-d",
        "--destination",
        type=str,
        default="Rotterdam Centraal",
        help="Destination station of the journey",
    )
    parser.add_argument(
        "-t", "--time", type=str, default="08:35:00", help="Departure time (hh:mm:ss)"
    )
    parser.add_argument(
        "-r",
        "--rounds",
        type=int,
        default=4,
        help="Number of rounds to execute the RAPTOR algorithm",
    )
    arguments = parser.parse_args()
    return arguments


def main(
    input_folder,
    output_folder,
    origin_station,
    destination_station,
    departure_time,
    rounds,
):
    """Run RAPTOR algorithm"""

    logger.debug("Input directory     : {}", input_folder)
    logger.debug("Output directory    : {}", output_folder)
    logger.debug("Origin station      : {}", origin_station)
    logger.debug("Destination station : {}", destination_station)
    logger.debug("Departure time      : {}", departure_time)
    logger.debug("Rounds              : {}", str(rounds))

    timetable = read_timetable(input_folder)

    logger.info("Calculating network from : {}".format(origin_station))

    # Departure time seconds
    dep_secs = str2sec(departure_time)
    logger.debug("Departure time (s.)  : " + str(dep_secs))

    # Find route between two stations
    start = perf_counter()
    bag_k, final_dest, evaluations = perform_raptor(
        timetable,
        origin_station,
        destination_station,
        dep_secs,
        rounds,
    )
    journey = reconstruct_journey(final_dest, bag=bag_k[rounds])
    print_journey(timetable, journey, dep_secs)

    logger.info(
        "RAPTOR Algorithm executed in {:.4f} seconds".format(perf_counter() - start)
    )

    if SAVE_RESULTS:
        write_results(output_folder, timetable, bag_k, evaluations)


def perform_raptor(
    timetable: Timetable,
    origin_station: str,
    destination_station: str,
    dep_secs: int,
    rounds: int,
):
    """
    Perform the Raptor algorithm.

    :param origin_station: Name of origin station
    :param destination_station: Name of destation station
    :param dep_secs: Time of departure in seconds
    :param rounds: Number of iterations to perform
    """

    start = perf_counter()

    from_stops = timetable.stations.get(origin_station).stops
    to_stops = timetable.stations.get(destination_station).stops

    # Initialize lookup with start node taking 0 seconds to reach
    # Bag contains per stop (travel_time, trip_id, previous_stop), trip_id is 0 in case of a transfer
    number_stops = len(timetable.stops) + 1
    bag = np.full(
        shape=(number_stops, 3),
        fill_value=(T24H, 0, -1),
        dtype=np.dtype(np.int32, np.int32, np.int32),
    )

    # Add origin stops to bag
    logger.debug("Starting from Stop IDs: {}".format(str(from_stops)))

    new_stops = []
    for from_stop in from_stops:
        bag[from_stop.index] = (0, 0, 0)
        new_stops.append(from_stop)

    # Run Round-Based Algorithm
    bag_k = {}
    evaluations = []

    for k in range(1, rounds + 1):

        logger.info("Analyzing possibilities round {}".format(k))

        # Get list of stops to evaluate in the process
        logger.debug("Stops to evaluate count: {}".format(len(new_stops)))

        # Update time to stops calculated based on stops reachable
        t = perf_counter()
        new_stops_travel, bag, evaluations = traverse_trips(
            timetable, evaluations, k, new_stops, bag, dep_secs
        )
        logger.debug(
            "Travel stops  calculated in {:0.4f} seconds".format(perf_counter() - t)
        )
        logger.debug("{} reachable stops added".format(len(new_stops_travel)))

        # Now add footpath transfers and update
        t = perf_counter()
        new_stops_transfer, bag = add_transfer_time(new_stops_travel, bag)
        logger.debug(
            "Transfers calculated in {:0.4f} seconds".format(perf_counter() - t)
        )
        logger.debug("{} transferable stops added".format(len(new_stops_transfer)))

        new_stops = set(new_stops_travel).union(new_stops_transfer)
        logger.debug("{} stops to evaluate in next round".format(len(new_stops)))

        # Store the results for this round
        bag_k[k] = np.copy(bag)

    # Determine the best destination ID, destination is a platform
    dest_stop = final_destination(to_stops, bag)
    if dest_stop != 0:
        logger.debug("Destination code   : {} ".format(dest_stop))
        logger.info(
            "Time to destination: {} minutes".format(bag[dest_stop.index][0] / 60)
        )
    else:
        logger.info("Destination unreachable with given parameters")

    logger.debug("Performing RAPTOR took {} seconds", perf_counter() - start)

    return bag_k, dest_stop, evaluations


if __name__ == "__main__":
    args = parse_arguments()
    main(
        args.input,
        args.output,
        args.origin,
        args.destination,
        args.time,
        args.rounds,
    )



"""Parse timetable from GTFS files"""
import os
import argparse
from typing import List
from dataclasses import dataclass
from collections import defaultdict

import pandas as pd
from loguru import logger

from pyraptor.dao.timetable import Timetable, write_timetable
from pyraptor.util import mkdir_if_not_exists, str2sec
from pyraptor.model.datatypes import (
    Stop,
    Stops,
    Trip,
    Trips,
    TripStopTime,
    TripStopTimes,
    Station,
    Stations,
)


@dataclass
class GtfsTimetable:
    """Gtfs Timetable data"""

    routes = None
    trips = None
    calendar = None
    stop_times = None
    stops = None


def parse_arguments():
    """Parse arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="data/NL-gtfs",
        help="Input directory",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="output/optimized_timetable",
        help="Input directory",
    )
    parser.add_argument(
        "-d", "--date", type=str, default="20210906", help="Departure date (yyyymmdd)"
    )
    parser.add_argument("-a", "--agencies", nargs="+", default=["NS"])
    arguments = parser.parse_args()
    return arguments


def main(
    input_folder: str, output_folder: str, departure_date: str, agencies: List[str]
):
    """Main function"""

    logger.info("Parse timetable from GTFS files")
    mkdir_if_not_exists(output_folder)

    gtfs_timetable = read_gtfs_timetable(input_folder, departure_date, agencies)
    timetable = optimize_timetable(gtfs_timetable)
    write_timetable(output_folder, timetable)


def read_gtfs_timetable(
    input_folder: str, departure_date: str, agencies: List[str]
) -> GtfsTimetable:
    """Extract operators from GTFS data"""

    logger.info("Reading GTFS data")

    # Read agencies
    logger.debug("Read Agencies")

    agencies_df = pd.read_csv(os.path.join(input_folder, "agency.txt"))
    agencies_df = agencies_df.loc[agencies_df["agency_name"].isin(agencies)][
        ["agency_id", "agency_name"]
    ]
    agency_ids = agencies_df.agency_id.values

    # Read routes
    logger.debug("Read Routes")

    routes = pd.read_csv(os.path.join(input_folder, "routes.txt"))
    routes = routes[routes.agency_id.isin(agency_ids)]
    routes = routes[
        ["route_id", "agency_id", "route_short_name", "route_long_name", "route_type"]
    ]

    # Read trips
    logger.debug("Read Trips")

    trips = pd.read_csv(os.path.join(input_folder, "trips.txt"))
    trips = trips[trips.route_id.isin(routes.route_id.values)]
    trips = trips[
        [
            "route_id",
            "service_id",
            "trip_id",
            "trip_short_name",
        ]
    ]
    trips["trip_short_name"] = trips["trip_short_name"].astype(int)

    # Read calendar
    logger.debug("Read Calendar")

    calendar = pd.read_csv(
        os.path.join(input_folder, "calendar_dates.txt"), dtype={"date": str}
    )
    calendar = calendar[calendar.service_id.isin(trips.service_id.values)]

    # Add date to trips and filter on departure date
    trips = trips.merge(calendar[["service_id", "date"]], on="service_id")
    trips = trips[trips.date == departure_date]

    # Read stop times
    logger.debug("Read Stop Times")

    stop_times = pd.read_csv(
        os.path.join(input_folder, "stop_times.txt"), dtype={"stop_id": str}
    )
    stop_times = stop_times[stop_times.trip_id.isin(trips.trip_id.values)]
    stop_times = stop_times[
        [
            "trip_id",
            "stop_sequence",
            "stop_id",
            "arrival_time",
            "departure_time",
        ]
    ]
    # Convert times to seconds
    stop_times["arrival_time"] = stop_times["arrival_time"].apply(str2sec)
    stop_times["departure_time"] = stop_times["departure_time"].apply(str2sec)

    # Read stops (platforms)
    logger.debug("Read Stops")

    stops_full = pd.read_csv(
        os.path.join(input_folder, "stops.txt"), dtype={"stop_id": str}
    )
    stops = stops_full.loc[stops_full["stop_id"].isin(stop_times.stop_id.unique())].copy()

    # Read stopareas, i.e. stations
    stopareas = stops["parent_station"].unique()
    stops = stops.append(stops_full.loc[stops_full["stop_id"].isin(stopareas)].copy())

    # stops["zone_id"] = stops["zone_id"].str.replace("IFF:", "").str.upper()
    stops["stop_code"] = stops.stop_code.str.upper()
    stops = stops[
        [
            "stop_id",
            "stop_name",
            "parent_station",
            "platform_code",
        ]
    ]

    # Filter out the general station codes
    stops = stops.loc[~stops.parent_station.isna()]

    logger.debug("Counts:")
    logger.debug("Routes     : {}", len(routes))
    logger.debug("Trips      : {}", len(trips))
    logger.debug("Stops      : {}", len(stops))
    logger.debug("Stop Times : {}", len(stop_times))

    gtfs_timetable = GtfsTimetable()
    gtfs_timetable.routes = routes
    gtfs_timetable.trips = trips
    gtfs_timetable.stop_times = stop_times
    gtfs_timetable.stops = stops

    return gtfs_timetable


def optimize_timetable(gtfs_timetable: GtfsTimetable) -> Timetable:
    """
    Optimize the timetable for usage in Raptor algorithm.
    """
    logger.info("Optimize GTFS timetable for algorithm")

    # Stations and stops, i.e. platforms
    stations = Stations()
    stops = Stops()

    for s in gtfs_timetable.stops.itertuples():
        # TODO parent_station instead of name as id
        station = Station(s.stop_name, s.stop_name)
        station = stations.add(station)

        stop = Stop(s.stop_id, s.stop_id, station, s.platform_code)

        station.add_stop(stop)
        stops.add(stop)

    # Stop Times
    stop_times = defaultdict(list)
    for stop_time in gtfs_timetable.stop_times.itertuples():
        stop_times[stop_time.trip_id].append(stop_time)

    # Trips and Trip Stop Times
    trips = Trips()
    trip_stop_times = TripStopTimes()

    for trip_row in gtfs_timetable.trips.itertuples():
        trip = Trip()
        trip.hint = trip_row.trip_short_name  # treinnummer

        # Iterate over stops
        sort_stop_times = sorted(
            stop_times[trip_row.trip_id], key=lambda s: int(s.stop_sequence)
        )
        for stopidx, stop_time in enumerate(sort_stop_times):
            # Timestamps
            dts_arr = stop_time.arrival_time
            dts_dep = stop_time.departure_time

            # Trip Stop Times
            stop = stops.get(stop_time.stop_id)
            trip_stop_time = TripStopTime(trip, stopidx, stop, dts_arr, dts_dep)

            trip_stop_times.add(trip_stop_time)
            trip.add_stop_time(trip_stop_time)

        # Add trip
        if trip:
            trips.add(trip)

    timetable = Timetable()
    timetable.stations = stations
    timetable.stops = stops
    timetable.trips = trips
    timetable.trip_stop_times = trip_stop_times

    return timetable


if __name__ == "__main__":
    args = parse_arguments()
    main(args.input, args.output, args.date, args.agencies)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyraptor', 'pyraptor.dao', 'pyraptor.gtfs', 'pyraptor.model']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.0.1,<2.0.0', 'loguru>=0.5.3,<0.6.0', 'pandas>=1.3.2,<2.0.0']

setup_kwargs = {
    'name': 'pyraptor',
    'version': '0.2.1',
    'description': 'Journey planner with RAPTOR algorithm',
    'long_description': '# PyRaptor\n\nPython implementation of RAPTOR using GTFS data.\n\nThree applications:\n\n- `pyraptor/gtfs/timetable.py` - Extract the time table information for one operator from a GTFS dataset and write it to an optimized format for querying with RAPTOR.\n- `pyraptor/query.py` - Get the best journey for a given origin, destination and desired departure time using RAPTOR\n- `pyraptor/range_query.py` - Get a list of the best journeys to all destinations for a given origin and desired departure time window using RAPTOR\n\n## Installation\n\nInstall from PyPi using `pip install pyraptor` or clone this repository and install from source using pip.\n\n## Example usage\n\n### 1. Create timetable from GTFS\n\n`python pyraptor/gtfs/timetable.py -d "20210223" -a NS`\n\n### 2. Run queries on timetable\n\n`python pyraptor/query.py -i output/optimized_timetable -or "Arnhem Zuid" -d "Oosterbeek" -t "08:30:00"`\n\n`python pyraptor/range_query.py -i output/optimized_timetable -or "Arnhem Zuid" -d "Oosterbeek" -st "08:00:00" -et "08:30:00"`\n\n# References\n\n[Round-Based Public Transit Routing](https://www.microsoft.com/en-us/research/wp-content/uploads/2012/01/raptor_alenex.pdf), Microsoft.com, Daniel Delling et al\n\n[Raptor, another journey planning algorithm](https://ljn.io/posts/raptor-journey-planning-algorithm), Linus Norton\n\n[Dutch GTFS feed](http://transitfeeds.com/p/ov/814), Transit Feeds\n',
    'author': 'Leo van der Meulen, Thom Hopmans',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lmeulen/pyraptor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
